#include "SoftwareSerial.h"

#define Start_Byte 0x7E
#define Version_Byte 0xFF
#define Command_Length 0x06
#define End_Byte 0xEF

// Returns info with command 0x41 [0x01: info, 0x00: no info]
#define Acknowledge 0x01

SoftwareSerial DFPlayer1(2, 3);  // RX, TX
SoftwareSerial DFPlayer2(4, 5);  // RX, TX
SoftwareSerial DFPlayer3(6, 7);  // RX, TX
byte receive_buffer[10] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
char data;                // Used for received commands from Serial Monitor
byte volume = 0x00;       // Used to store current volume level
bool mute_state = false;  // Used to toggle mute state
// Excecute the command and parameters

void execute_CMD(SoftwareSerial mySerial, byte Command, byte Data1, byte Data2) {
  // Calculate the checksum (2 bytes)
  word Checksum = -(Version_Byte + Command_Length + Command + Acknowledge + Data1 + Data2);
  // Build the command
  byte command_line[10] = { Start_Byte, Version_Byte,
                            Command_Length, Command, Acknowledge,
                            Data1, Data2, highByte(Checksum),
                            lowByte(Checksum), End_Byte };
  // Send the command line to the module
  for (byte k = 0; k < 10; k++) {
    mySerial.write(command_line[k]);
  }
}

void reset_rec_buf() {
  for (uint8_t i = 0; i < 10; i++) {
    receive_buffer[i] = 0;
  }
}

bool receive(SoftwareSerial mySerial) {
  reset_rec_buf();
  if (mySerial.available() < 10) {
    return false;
  }
  for (uint8_t i = 0; i < 10; i++) {
    short b = mySerial.read();
    if (b == -1) {
      return false;
    }
    receive_buffer[i] = b;
  }
  // When you reset the module in software,
  // received buffer elements are shifted.
  // To correct that we do the following:
  short b = receive_buffer[0];
  for (uint8_t i = 0; i < 10; i++) {
    if (i == 9) {
      receive_buffer[i] = b;
    } else {
      receive_buffer[i] = receive_buffer[i + 1];
    }
  }  // End correcting receive_buffer
  return true;
}

void print_received(bool print_it) {
  if (print_it) {
    if (receive(DFPlayer1)) {
      for (uint8_t i = 0; i < 10; i++) {
        Serial.print(receive_buffer[i], HEX);
        Serial.print("\t");
      }
      Serial.println();
    }
    if (receive(DFPlayer2)) {
      for (uint8_t i = 0; i < 10; i++) {
        Serial.print(receive_buffer[i], HEX);
        Serial.print("\t");
      }
      Serial.println();
    }
  } else {
    receive(DFPlayer1);
    receive(DFPlayer2);
  }
}
void module_init() {
  execute_CMD(DFPlayer1,0x0C, 0, 0);
  execute_CMD(DFPlayer2,0x0C, 0, 0);
  execute_CMD(DFPlayer3,0x0C, 0, 0);
  delay(1000);  // Reset the module
  print_received(false);
  delay(100);
  Serial.print("SDON\t");
  print_received(true);
  delay(100);
  play_first();
  set_volume(0x09);
}
void play_first() {
  Serial.print("PLYFST\t");
  execute_CMD(DFPlayer1,0x03, 0, 1);
  execute_CMD(DFPlayer2,0x03, 0, 1);
  execute_CMD(DFPlayer3,0x03, 0, 1);
  delay(100);  // Play first song
  print_received(false);
  delay(100);
  execute_CMD(DFPlayer1,0x45, 0, 0);
  execute_CMD(DFPlayer2,0x45, 0, 0);
  execute_CMD(DFPlayer3,0x45, 0, 0);
  delay(100);  // Get playback status
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void set_volume(uint8_t volume) {
  Serial.print("SETVOL\t");
  execute_CMD(DFPlayer1,0x06, 0, volume);
  execute_CMD(DFPlayer2,0x06, 0, volume);
  execute_CMD(DFPlayer3,0x06, 0, volume);
  delay(100);  // Set volume level
  print_received(false);
  delay(100);
  execute_CMD(DFPlayer1,0x43, 0, 0);
  execute_CMD(DFPlayer2,0x43, 0, 0);
  execute_CMD(DFPlayer3,0x43, 0, 0);
  delay(100);  // Get volume level
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void play_next() {
  Serial.print("NEXT\t");
  execute_CMD(DFPlayer1,0x01, 0, 0);
  execute_CMD(DFPlayer2,0x01, 0, 0);
  execute_CMD(DFPlayer3,0x01, 0, 0);
  delay(100);
  print_received(false);
  delay(100);
  execute_CMD(DFPlayer1,0x4C, 0, 0);
  execute_CMD(DFPlayer2,0x4C, 0, 0);
  execute_CMD(DFPlayer3,0x4C, 0, 0);
  delay(100);  // Get current song played
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void mute() {
  mute_state = !mute_state;
  if (mute_state) {
    execute_CMD(DFPlayer1,0x43, 0, 0);
    execute_CMD(DFPlayer2,0x43, 0, 0);
    execute_CMD(DFPlayer3,0x43, 0, 0);
    delay(100);  // Return volume leve
    print_received(false);
    delay(100);
    print_received(false);
    delay(100);
    volume = receive_buffer[6];
    Serial.print("MUTE\t");
    execute_CMD(DFPlayer1,0x06, 0, 0x00);
    execute_CMD(DFPlayer2,0x06, 0, 0x00);
    execute_CMD(DFPlayer3,0x06, 0, 0x00);
    delay(100);  // Set volume level
    print_received(false);
    delay(100);
    execute_CMD(DFPlayer1,0x43, 0, 0);
    execute_CMD(DFPlayer2,0x43, 0, 0);
    execute_CMD(DFPlayer3,0x43, 0, 0);
    delay(100);  // Get volume level
    print_received(false);
    delay(100);
    print_received(true);
    delay(100);
  }
  // one tab
  else {
    Serial.print("VOL\t");
    execute_CMD(DFPlayer1,0x06, 0, volume);
    execute_CMD(DFPlayer2,0x06, 0, volume);
    execute_CMD(DFPlayer3,0x06, 0, volume);
    delay(100);  // Set previous vol
    print_received(false);
    delay(100);
    execute_CMD(DFPlayer1,0x43, 0, 0);
    execute_CMD(DFPlayer2,0x43, 0, 0);
    execute_CMD(DFPlayer3,0x43, 0, 0);
    delay(100);  // Get volume level
    print_received(false);
    delay(100);
    print_received(true);
    delay(100);
  }
}

void random_play() {
  // Random plays all songs, loops all, repeats songs in playback
  execute_CMD(DFPlayer1,0x18, 0, 0);
  execute_CMD(DFPlayer2,0x18, 0, 0);
  execute_CMD(DFPlayer3,0x18, 0, 0);
  delay(100);
  Serial.print("RNDM\t");
  print_received(false);
  delay(100);
  execute_CMD(DFPlayer1,0x4C, 0, 0);  // Get current song played
  execute_CMD(DFPlayer2,0x4C, 0, 0);  // Get current song played
  execute_CMD(DFPlayer3,0x4C, 0, 0);  // Get current song played
  delay(100);
  print_received(false);
  delay(100);
  print_received(true);
  delay(100);
}

void query_status() {
  execute_CMD(DFPlayer1,0x42, 0, 0);
  execute_CMD(DFPlayer2,0x42, 0, 0);
  execute_CMD(DFPlayer3,0x42, 0, 0);
  delay(100);  // Get status of module
  print_received(false);
  delay(100);
  Serial.print("STATUS\t");
  print_received(true);
  delay(100);
  execute_CMD(DFPlayer1,0x43, 0, 0);
  execute_CMD(DFPlayer2,0x43, 0, 0);
  execute_CMD(DFPlayer3,0x43, 0, 0);
  delay(100);  // Get volume level
  print_received(false);
  delay(100);
  Serial.print("VOLUME\t");
  print_received(true);
  delay(100);
  execute_CMD(DFPlayer1,0x44, 0, 0);
  execute_CMD(DFPlayer2,0x44, 0, 0);
  execute_CMD(DFPlayer3,0x44, 0, 0);
  delay(100);  // Get EQ status
  print_received(false);
  delay(100);
  Serial.print("EQ\t");
  print_received(true);
  delay(100);
  execute_CMD(DFPlayer1,0x45, 0, 0);
  execute_CMD(DFPlayer2,0x45, 0, 0);
  execute_CMD(DFPlayer3,0x45, 0, 0);
  delay(100);  // Get playback status
  print_received(false);
  delay(100);
  Serial.print("PLYBCK\t");
  print_received(true);
  delay(100);
  execute_CMD(DFPlayer1,0x46, 0, 0);
  execute_CMD(DFPlayer2,0x46, 0, 0);
  execute_CMD(DFPlayer3,0x46, 0, 0);
  delay(100);  // Get software version
  print_received(false);
  delay(100);
  Serial.print("SFVER\t");
  print_received(true);
  delay(100);
  // Get total number of files on storage device
  execute_CMD(DFPlayer1,0x48, 0, 0);
  execute_CMD(DFPlayer2,0x48, 0, 0);
  execute_CMD(DFPlayer3,0x48, 0, 0);
  delay(100);
  print_received(false);
  delay(100);
  Serial.print("FILES\t");
  print_received(true);
  delay(100);
  execute_CMD(DFPlayer1,0x4C, 0, 0);
  execute_CMD(DFPlayer2,0x4C, 0, 0);
  execute_CMD(DFPlayer3,0x4C, 0, 0);
  delay(100);  // Get current song played
  print_received(false);
  delay(100);
  Serial.print("CRRTRK\t");
  print_received(true);
  delay(100);
}

void setup() {
  Serial.begin(115200);
  DFPlayer1.begin(9600);
  DFPlayer2.begin(9600);
  DFPlayer3.begin(9600);
  delay(1000);
  Serial.println("\nInitialization");
  module_init();
}
void loop() {
  print_received(true);
  while (Serial.available() > 0) {
    data = Serial.read();
    // Serial.println(data, HEX); // For debugging
    if (data != "/n") {
      if (data == 'N') {
        Serial.println("\nPlay next song");
        play_next();
      } else if (data == 'B') {
        Serial.println("\nRandom play");
        random_play();
      }
      // .....
      else if (data == 'D') {
        Serial.println("\nQuerry status of the module");
        query_status();
      }
    }
  }
  delay(100);
}